@@ -28,6 +28,7 @@
 
 #include <glib.h>
 #include <gatchat.h>
+#include <gattty.h>
 
 #define OFONO_API_SUBJECT_TO_CHANGE
 #include <ofono/plugin.h>
@@ -49,90 +50,6 @@
 
 #include <drivers/atmodem/vendor.h>
 
-/* Supply our own syntax parser */
-
-enum G1_STATE_ {
-	G1_STATE_IDLE = 0,
-	G1_STATE_RESPONSE,
-	G1_STATE_GUESS_PDU,
-	G1_STATE_PDU,
-	G1_STATE_PROMPT,
-};
-
-static void g1_hint(GAtSyntax *syntax, GAtSyntaxExpectHint hint)
-{
-	if (hint == G_AT_SYNTAX_EXPECT_PDU)
-		syntax->state = G1_STATE_GUESS_PDU;
-}
-
-static GAtSyntaxResult g1_feed(GAtSyntax *syntax,
-		const char *bytes, gsize *len)
-{
-	gsize i = 0;
-	GAtSyntaxResult res = G_AT_SYNTAX_RESULT_UNSURE;
-
-	while (i < *len) {
-		char byte = bytes[i];
-
-		switch (syntax->state) {
-		case G1_STATE_IDLE:
-			if (byte == '\r' || byte == '\n')
-				/* ignore */;
-			else if (byte == '>')
-				syntax->state = G1_STATE_PROMPT;
-			else
-				syntax->state = G1_STATE_RESPONSE;
-			break;
-
-		case G1_STATE_RESPONSE:
-			if (byte == '\r') {
-				syntax->state = G1_STATE_IDLE;
-
-				i += 1;
-				res = G_AT_SYNTAX_RESULT_LINE;
-				goto out;
-			}
-			break;
-
-		case G1_STATE_GUESS_PDU:
-			/* keep going until we find a LF that leads the PDU */
-			if (byte == '\n')
-				syntax->state = G1_STATE_PDU;
-			break;
-
-		case G1_STATE_PDU:
-			if (byte == '\r') {
-				syntax->state = G1_STATE_IDLE;
-
-				i += 1;
-				res = G_AT_SYNTAX_RESULT_PDU;
-				goto out;
-			}
-			break;
-
-		case G1_STATE_PROMPT:
-			if (byte == ' ') {
-				syntax->state = G1_STATE_IDLE;
-				i += 1;
-				res = G_AT_SYNTAX_RESULT_PROMPT;
-				goto out;
-			}
-
-			syntax->state = G1_STATE_RESPONSE;
-			return G_AT_SYNTAX_RESULT_UNSURE;
-
-		default:
-			break;
-		};
-
-		i += 1;
-	}
-
-out:
-	*len = i;
-	return res;
-}
-
 static void g1_debug(const char *str, void *data)
 {
 	DBG("%s", str);
@@ -141,39 +58,14 @@ static void g1_debug(const char *str, void *data)
 /* Detect hardware, and initialize if found */
 static int g1_probe(struct ofono_modem *modem)
 {
-	GAtSyntax *syntax;
-	GAtChat *chat;
-	const char *device;
-
 	DBG("");
 
-	device = ofono_modem_get_string(modem, "Device");
-	if (device == NULL)
-		return -EINVAL;
-
-	syntax = g_at_syntax_new_full(g1_feed, g1_hint, G1_STATE_IDLE);
-	chat = g_at_chat_new_from_tty(device, syntax);
-	g_at_syntax_unref(syntax);
-
-	if (chat == NULL)
-		return -EIO;
-
-	if (getenv("OFONO_AT_DEBUG") != NULL)
-		g_at_chat_set_debug(chat, g1_debug, NULL);
-
-	ofono_modem_set_data(modem, chat);
-
 	return 0;
 }
 
 static void g1_remove(struct ofono_modem *modem)
 {
-	GAtChat *chat = ofono_modem_get_data(modem);
-
 	DBG("");
-
-	ofono_modem_set_data(modem, NULL);
-	g_at_chat_unref(chat);
 }
 
 static void cfun_set_on_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -189,10 +81,34 @@ static void cfun_set_on_cb(gboolean ok, GAtResult *result, gpointer user_data)
 /* power up hardware */
 static int g1_enable(struct ofono_modem *modem)
 {
-	GAtChat *chat = ofono_modem_get_data(modem);
+	GAtSyntax *syntax;
+	GIOChannel *channel;
+	GAtChat *chat;
+	const char *device;
 
 	DBG("");
 
+	device = ofono_modem_get_string(modem, "Device");
+	if (device == NULL)
+		return -EINVAL;
+
+	channel = g_at_tty_open(device, NULL);
+	if (channel == NULL)
+		return -EIO;
+
+	syntax = g_at_syntax_new_gsm_permissive();
+	chat = g_at_chat_new(channel, syntax);
+	g_io_channel_unref(channel);
+	g_at_syntax_unref(syntax);
+
+	if (chat == NULL)
+		return -EIO;
+
+	if (getenv("OFONO_AT_DEBUG") != NULL)
+		g_at_chat_set_debug(chat, g1_debug, NULL);
+
+	ofono_modem_set_data(modem, chat);
+
 	/* ensure modem is in a known state; verbose on, echo/quiet off */
 	g_at_chat_send(chat, "ATE0Q0V1", NULL, NULL, NULL, NULL);
 
@@ -221,23 +137,34 @@ static int g1_disable(struct ofono_modem *modem)
 	/* power down modem */
 	g_at_chat_send(chat, "AT+CFUN=0", NULL, cfun_set_off_cb, modem, NULL);
 
+	g_at_chat_unref(chat);
+	ofono_modem_set_data(modem, NULL);
+
 	return 0;
 }
 
-static void g1_populate(struct ofono_modem *modem)
+static void g1_pre_sim(struct ofono_modem *modem)
+{
+	GAtChat *chat = ofono_modem_get_data(modem);
+
+	DBG("");
+
+	ofono_devinfo_create(modem, 0, "atmodem", chat);
+	ofono_sim_create(modem, 0, "atmodem", chat);
+	ofono_voicecall_create(modem, 0, "atmodem", chat);
+}
+
+static void g1_post_sim(struct ofono_modem *modem)
 {
 	GAtChat *chat = ofono_modem_get_data(modem);
 	struct ofono_message_waiting *mw;
 
 	DBG("");
 
-	ofono_devinfo_create(modem, 0, "atmodem", chat);
 	ofono_ussd_create(modem, 0, "atmodem", chat);
-	ofono_sim_create(modem, 0, "atmodem", chat);
 	ofono_call_forwarding_create(modem, 0, "atmodem", chat);
 	ofono_call_settings_create(modem, 0, "atmodem", chat);
 	ofono_netreg_create(modem, 0, "atmodem", chat);
-	ofono_voicecall_create(modem, 0, "atmodem", chat);
 	ofono_call_meter_create(modem, 0, "atmodem", chat);
 	ofono_call_barring_create(modem, 0, "atmodem", chat);
 	ofono_ssn_create(modem, 0, "atmodem", chat);
@@ -255,7 +182,8 @@ static struct ofono_modem_driver g1_driver = {
 	.remove		= g1_remove,
 	.enable		= g1_enable,
 	.disable	= g1_disable,
-	.populate	= g1_populate,
+	.pre_sim	= g1_pre_sim,
+	.post_sim	= g1_post_sim,
 };
 
 static int g1_init(void)
