@@ -26,241 +26,167 @@
 #define _GNU_SOURCE
 #include <string.h>
 #include <stdio.h>
+#include <errno.h>
 
 #include <glib.h>
 #include <gdbus.h>
 
 #include "ofono.h"
 
-#include "driver.h"
 #include "common.h"
-#include "ussd.h"
 
 #define SUPPLEMENTARY_SERVICES_INTERFACE "org.ofono.SupplementaryServices"
 
 #define USSD_FLAG_PENDING 0x1
 
+static GSList *g_drivers = NULL;
+
+enum ussd_status {
+	USSD_STATUS_NOTIFY = 0,
+	USSD_STATUS_ACTION_REQUIRED = 1,
+	USSD_STATUS_TERMINATED = 2,
+	USSD_STATUS_LOCAL_CLIENT_RESPONDED = 3,
+	USSD_STATUS_NOT_SUPPORTED = 4,
+	USSD_STATUS_TIMED_OUT = 5,
+};
+
 enum ussd_state {
 	USSD_STATE_IDLE = 0,
 	USSD_STATE_ACTIVE = 1,
 	USSD_STATE_USER_ACTION = 2
 };
 
-struct ussd_data {
-	struct ofono_ussd_ops *ops;
+struct ofono_ussd {
 	int state;
 	DBusMessage *pending;
 	int flags;
+	GSList *ss_control_list;
+	GSList *ss_passwd_list;
+	const struct ofono_ussd_driver *driver;
+	void *driver_data;
+	struct ofono_atom *atom;
 };
 
-static struct ussd_data *ussd_create()
-{
-	struct ussd_data *r;
-
-	r = g_try_new0(struct ussd_data, 1);
-
-	return r;
-}
-
-static void ussd_destroy(gpointer data)
-{
-	struct ofono_modem *modem = data;
-	struct ussd_data *ussd = modem->ussd;
-
-	g_free(ussd);
-}
-
-struct ss_control_entry {
+struct ssc_entry {
 	char *service;
-	ss_control_cb_t cb;
+	void *cb;
+	void *user;
+	ofono_destroy_func destroy;
 };
 
-static struct ss_control_entry *ss_control_entry_create(const char *service,
-							ss_control_cb_t cb)
+static struct ssc_entry *ssc_entry_create(const char *sc, void *cb, void *data,
+						ofono_destroy_func destroy)
 {
-	struct ss_control_entry *r;
+	struct ssc_entry *r;
 
-	r = g_try_new0(struct ss_control_entry, 1);
+	r = g_try_new0(struct ssc_entry, 1);
 
 	if (!r)
 		return r;
 
-	r->service = g_strdup(service);
+	r->service = g_strdup(sc);
 	r->cb = cb;
+	r->user = data;
+	r->destroy = destroy;
 
 	return r;
 }
 
-static void ss_control_entry_destroy(struct ss_control_entry *ca)
+static void ssc_entry_destroy(struct ssc_entry *ca)
 {
+	if (ca->destroy)
+		ca->destroy(ca->user);
+
 	g_free(ca->service);
 	g_free(ca);
 }
 
-static gint ss_control_entry_compare(gconstpointer a, gconstpointer b)
+static gint ssc_entry_find_by_service(gconstpointer a, gconstpointer b)
 {
-	const struct ss_control_entry *ca = a;
-	const struct ss_control_entry *cb = b;
-	int ret;
-
-	ret = strcmp(ca->service, cb->service);
-
-	if (ret)
-		return ret;
-
-	if (ca->cb < cb->cb)
-		return -1;
-
-	if (ca->cb > cb->cb)
-		return 1;
-
-	return 0;
-}
-
-static gint ss_control_entry_find_by_service(gconstpointer a, gconstpointer b)
-{
-	const struct ss_control_entry *ca = a;
+	const struct ssc_entry *ca = a;
 
 	return strcmp(ca->service, b);
 }
 
-gboolean ss_control_register(struct ofono_modem *modem, const char *str,
-				ss_control_cb_t cb)
+gboolean __ofono_ussd_ssc_register(struct ofono_ussd *ussd, const char *sc,
+					ofono_ussd_ssc_cb_t cb, void *data,
+					ofono_destroy_func destroy)
 {
-	struct ss_control_entry *entry;
+	struct ssc_entry *entry;
 
-	if (!modem)
+	if (!ussd)
 		return FALSE;
 
-	entry = ss_control_entry_create(str, cb);
+	entry = ssc_entry_create(sc, cb, data, destroy);
 
 	if (!entry)
 		return FALSE;
 
-	modem->ss_control_list = g_slist_append(modem->ss_control_list, entry);
+	ussd->ss_control_list = g_slist_prepend(ussd->ss_control_list, entry);
 
 	return TRUE;
 }
 
-void ss_control_unregister(struct ofono_modem *modem, const char *str,
-				ss_control_cb_t cb)
+void __ofono_ussd_ssc_unregister(struct ofono_ussd *ussd, const char *sc)
 {
-	const struct ss_control_entry entry = { (char *)str, cb };
 	GSList *l;
 
-	if (!modem)
+	if (!ussd)
 		return;
 
-	l = g_slist_find_custom(modem->ss_control_list, &entry,
-				ss_control_entry_compare);
+	l = g_slist_find_custom(ussd->ss_control_list, sc,
+				ssc_entry_find_by_service);
 
 	if (!l)
 		return;
 
-	ss_control_entry_destroy(l->data);
-	modem->ss_control_list = g_slist_remove(modem->ss_control_list,
-						l->data);
+	ssc_entry_destroy(l->data);
+	ussd->ss_control_list = g_slist_remove(ussd->ss_control_list, l->data);
 }
 
-struct ss_passwd_entry {
-	char *service;
-	ss_passwd_cb_t cb;
-};
-
-static struct ss_passwd_entry *ss_passwd_entry_create(const char *service,
-							ss_passwd_cb_t cb)
+gboolean __ofono_ussd_passwd_register(struct ofono_ussd *ussd, const char *sc,
+					ofono_ussd_passwd_cb_t cb, void *data,
+					ofono_destroy_func destroy)
 {
-	struct ss_passwd_entry *r;
+	struct ssc_entry *entry;
 
-	r = g_try_new0(struct ss_passwd_entry, 1);
-
-	if (!r)
-		return r;
-
-	r->service = g_strdup(service);
-	r->cb = cb;
-
-	return r;
-}
-
-static void ss_passwd_entry_destroy(struct ss_passwd_entry *ca)
-{
-	g_free(ca->service);
-	g_free(ca);
-}
-
-static gint ss_passwd_entry_compare(gconstpointer a, gconstpointer b)
-{
-	const struct ss_passwd_entry *ca = a;
-	const struct ss_passwd_entry *cb = b;
-	int ret;
-
-	ret = strcmp(ca->service, cb->service);
-
-	if (ret)
-		return ret;
-
-	if (ca->cb < cb->cb)
-		return -1;
-
-	if (ca->cb > cb->cb)
-		return 1;
-
-	return 0;
-}
-
-static gint ss_passwd_entry_find_by_service(gconstpointer a, gconstpointer b)
-{
-	const struct ss_passwd_entry *ca = a;
-
-	return strcmp(ca->service, b);
-}
-
-gboolean ss_passwd_register(struct ofono_modem *modem, const char *str,
-				ss_passwd_cb_t cb)
-{
-	struct ss_passwd_entry *entry;
-
-	if (!modem)
+	if (!ussd)
 		return FALSE;
 
-	entry = ss_passwd_entry_create(str, cb);
+	entry = ssc_entry_create(sc, cb, data, destroy);
 
 	if (!entry)
 		return FALSE;
 
-	modem->ss_passwd_list = g_slist_append(modem->ss_passwd_list, entry);
+	ussd->ss_passwd_list = g_slist_prepend(ussd->ss_passwd_list, entry);
 
 	return TRUE;
 }
 
-void ss_passwd_unregister(struct ofono_modem *modem, const char *str,
-				ss_passwd_cb_t cb)
+void __ofono_ussd_passwd_unregister(struct ofono_ussd *ussd, const char *sc)
 {
-	const struct ss_passwd_entry entry = { (char *)str, cb };
 	GSList *l;
 
-	if (!modem)
+	if (!ussd)
 		return;
 
-	l = g_slist_find_custom(modem->ss_passwd_list, &entry,
-				ss_passwd_entry_compare);
+	l = g_slist_find_custom(ussd->ss_passwd_list, sc,
+				ssc_entry_find_by_service);
 
 	if (!l)
 		return;
 
-	ss_passwd_entry_destroy(l->data);
-	modem->ss_passwd_list = g_slist_remove(modem->ss_passwd_list,
-						l->data);
+	ssc_entry_destroy(l->data);
+	ussd->ss_passwd_list = g_slist_remove(ussd->ss_passwd_list, l->data);
 }
 
-static gboolean recognized_passwd_change_string(struct ofono_modem *modem,
+static gboolean recognized_passwd_change_string(struct ofono_ussd *ussd,
 						int type, char *sc,
 						char *sia, char *sib,
 						char *sic, char *sid,
 						char *dn, DBusMessage *msg)
 {
-	GSList *l = modem->ss_passwd_list;
+	GSList *l = ussd->ss_passwd_list;
 
 	switch (type) {
 	case SS_CONTROL_TYPE_ACTIVATION:
@@ -283,10 +209,11 @@ static gboolean recognized_passwd_change_string(struct ofono_modem *modem,
 	}
 
 	while ((l = g_slist_find_custom(l, sia,
-			ss_passwd_entry_find_by_service)) != NULL) {
-		struct ss_passwd_entry *entry = l->data;
+			ssc_entry_find_by_service)) != NULL) {
+		struct ssc_entry *entry = l->data;
+		ofono_ussd_passwd_cb_t cb = entry->cb;
 
-		if (entry->cb(modem, sia, sib, sic, msg))
+		if (cb(sia, sib, sic, msg, entry->user))
 			return TRUE;
 
 		l = l->next;
@@ -295,7 +222,7 @@ static gboolean recognized_passwd_change_string(struct ofono_modem *modem,
 	return FALSE;
 }
 
-static gboolean recognized_control_string(struct ofono_modem *modem,
+static gboolean recognized_control_string(struct ofono_ussd *ussd,
 						const char *ss_str,
 						DBusMessage *msg)
 {
@@ -308,7 +235,7 @@ static gboolean recognized_control_string(struct ofono_modem *modem,
 
 	if (parse_ss_control_string(str, &type, &sc,
 				&sia, &sib, &sic, &sid, &dn)) {
-		GSList *l = modem->ss_control_list;
+		GSList *l = ussd->ss_control_list;
 
 		ofono_debug("Got parse result: %d, %s, %s, %s, %s, %s, %s",
 				type, sc, sia, sib, sic, sid, dn);
@@ -316,7 +243,7 @@ static gboolean recognized_control_string(struct ofono_modem *modem,
 		/* A password change string needs to be treated separately
 		 * because it uses a fourth SI and is thus not a valid
 		 * control string.  */
-		if (recognized_passwd_change_string(modem, type, sc,
+		if (recognized_passwd_change_string(ussd, type, sc,
 					sia, sib, sic, sid, dn, msg)) {
 			ret = TRUE;
 			goto out;
@@ -326,10 +253,11 @@ static gboolean recognized_control_string(struct ofono_modem *modem,
 			goto out;
 
 		while ((l = g_slist_find_custom(l, sc,
-				ss_control_entry_find_by_service)) != NULL) {
-			struct ss_control_entry *entry = l->data;
+				ssc_entry_find_by_service)) != NULL) {
+			struct ssc_entry *entry = l->data;
+			ofono_ussd_ssc_cb_t cb = entry->cb;
 
-			if (entry->cb(modem, type, sc, sia, sib, sic, dn, msg)) {
+			if (cb(type, sc, sia, sib, sic, dn, msg, entry->user)) {
 				ret = TRUE;
 				goto out;
 			}
@@ -355,9 +283,8 @@ out:
 	return ret;
 }
 
-void ofono_ussd_notify(struct ofono_modem *modem, int status, const char *str)
+void ofono_ussd_notify(struct ofono_ussd *ussd, int status, const char *str)
 {
-	struct ussd_data *ussd = modem->ussd;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *ussdstr = "USSD";
 	const char sig[] = { DBUS_TYPE_STRING, 0 };
@@ -416,7 +343,7 @@ out:
 
 static void ussd_callback(const struct ofono_error *error, void *data)
 {
-	struct ussd_data *ussd = data;
+	struct ofono_ussd *ussd = data;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	DBusMessage *reply;
 
@@ -445,8 +372,7 @@ static void ussd_callback(const struct ofono_error *error, void *data)
 static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct ussd_data *ussd = modem->ussd;
+	struct ofono_ussd *ussd = data;
 	const char *str;
 
 	if (ussd->flags & USSD_FLAG_PENDING)
@@ -463,7 +389,7 @@ static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 		return __ofono_error_invalid_format(msg);
 
 	ofono_debug("checking if this is a recognized control string");
-	if (recognized_control_string(modem, str, msg))
+	if (recognized_control_string(ussd, str, msg))
 		return NULL;
 
 	ofono_debug("No.., checking if this is a USSD string");
@@ -472,20 +398,20 @@ static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 
 	ofono_debug("OK, running USSD request");
 
-	if (!ussd->ops->request)
+	if (!ussd->driver->request)
 		return __ofono_error_not_implemented(msg);
 
 	ussd->flags |= USSD_FLAG_PENDING;
 	ussd->pending = dbus_message_ref(msg);
 
-	ussd->ops->request(modem, str, ussd_callback, ussd);
+	ussd->driver->request(ussd, str, ussd_callback, ussd);
 
 	return NULL;
 }
 
 static void ussd_cancel_callback(const struct ofono_error *error, void *data)
 {
-	struct ussd_data *ussd = data;
+	struct ofono_ussd *ussd = data;
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
@@ -510,8 +436,7 @@ static void ussd_cancel_callback(const struct ofono_error *error, void *data)
 static DBusMessage *ussd_cancel(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct ussd_data *ussd = modem->ussd;
+	struct ofono_ussd *ussd = data;
 
 	if (ussd->flags & USSD_FLAG_PENDING)
 		return __ofono_error_busy(msg);
@@ -519,13 +444,13 @@ static DBusMessage *ussd_cancel(DBusConnection *conn, DBusMessage *msg,
 	if (ussd->state == USSD_STATE_IDLE)
 		return __ofono_error_not_active(msg);
 
-	if (!ussd->ops->cancel)
+	if (!ussd->driver->cancel)
 		return __ofono_error_not_implemented(msg);
 
 	ussd->flags |= USSD_FLAG_PENDING;
 	ussd->pending = dbus_message_ref(msg);
 
-	ussd->ops->cancel(modem, ussd_cancel_callback, ussd);
+	ussd->driver->cancel(ussd, ussd_cancel_callback, ussd);
 
 	return NULL;
 }
@@ -542,48 +467,127 @@ static GDBusSignalTable ussd_signals[] = {
 	{ }
 };
 
-int ofono_ussd_register(struct ofono_modem *modem, struct ofono_ussd_ops *ops)
+int ofono_ussd_driver_register(const struct ofono_ussd_driver *d)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (modem == NULL)
-		return -1;
+	if (d->probe == NULL)
+		return -EINVAL;
 
-	if (ops == NULL)
-		return -1;
-
-	modem->ussd = ussd_create();
-
-	if (modem->ussd == NULL)
-		return -1;
-
-	modem->ussd->ops = ops;
-
-	if (!g_dbus_register_interface(conn, modem->path,
-					SUPPLEMENTARY_SERVICES_INTERFACE,
-					ussd_methods, ussd_signals, NULL,
-					modem, ussd_destroy)) {
-		ofono_error("Could not create %s interface",
-				SUPPLEMENTARY_SERVICES_INTERFACE);
-
-		ussd_destroy(modem->ussd);
-
-		return -1;
-	}
-
-	ofono_modem_add_interface(modem, SUPPLEMENTARY_SERVICES_INTERFACE);
+	g_drivers = g_slist_prepend(g_drivers, (void *)d);
 
 	return 0;
 }
 
-void ofono_ussd_unregister(struct ofono_modem *modem)
+void ofono_ussd_driver_unregister(const struct ofono_ussd_driver *d)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (modem->ussd == NULL)
-		return;
+	g_drivers = g_slist_remove(g_drivers, (void *)d);
+}
+
+static void ussd_unregister(struct ofono_atom *atom)
+{
+	struct ofono_ussd *ussd = __ofono_atom_get_data(atom);
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
+	const char *path = __ofono_atom_get_path(atom);
+
+	g_slist_foreach(ussd->ss_control_list, (GFunc)ssc_entry_destroy, NULL);
+	g_slist_free(ussd->ss_control_list);
+	ussd->ss_control_list = NULL;
+
+	g_slist_foreach(ussd->ss_passwd_list, (GFunc)ssc_entry_destroy, NULL);
+	g_slist_free(ussd->ss_passwd_list);
+	ussd->ss_passwd_list = NULL;
 
 	ofono_modem_remove_interface(modem, SUPPLEMENTARY_SERVICES_INTERFACE);
-	g_dbus_unregister_interface(conn, modem->path,
+	g_dbus_unregister_interface(conn, path,
 					SUPPLEMENTARY_SERVICES_INTERFACE);
 }
+
+static void ussd_remove(struct ofono_atom *atom)
+{
+	struct ofono_ussd *ussd = __ofono_atom_get_data(atom);
+
+	DBG("atom: %p", atom);
+
+	if (ussd == NULL)
+		return;
+
+	if (ussd->driver && ussd->driver->remove)
+		ussd->driver->remove(ussd);
+
+	g_free(ussd);
+}
+
+struct ofono_ussd *ofono_ussd_create(struct ofono_modem *modem,
+					unsigned int vendor,
+					const char *driver,
+					void *data)
+{
+	struct ofono_ussd *ussd;
+	GSList *l;
+
+	if (driver == NULL)
+		return NULL;
+
+	ussd = g_try_new0(struct ofono_ussd, 1);
+
+	if (ussd == NULL)
+		return NULL;
+
+	ussd->atom = __ofono_modem_add_atom(modem, OFONO_ATOM_TYPE_USSD,
+						ussd_remove, ussd);
+
+	for (l = g_drivers; l; l = l->next) {
+		const struct ofono_ussd_driver *drv = l->data;
+
+		if (g_strcmp0(drv->name, driver))
+			continue;
+
+		if (drv->probe(ussd, vendor, data) < 0)
+			continue;
+
+		ussd->driver = drv;
+		break;
+	}
+
+	return ussd;
+}
+
+void ofono_ussd_register(struct ofono_ussd *ussd)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(ussd->atom);
+	const char *path = __ofono_atom_get_path(ussd->atom);
+
+	if (!g_dbus_register_interface(conn, path,
+					SUPPLEMENTARY_SERVICES_INTERFACE,
+					ussd_methods, ussd_signals, NULL,
+					ussd, NULL)) {
+		ofono_error("Could not create %s interface",
+				SUPPLEMENTARY_SERVICES_INTERFACE);
+
+		return;
+	}
+
+	ofono_modem_add_interface(modem, SUPPLEMENTARY_SERVICES_INTERFACE);
+
+	__ofono_atom_register(ussd->atom, ussd_unregister);
+}
+
+void ofono_ussd_remove(struct ofono_ussd *ussd)
+{
+	__ofono_atom_free(ussd->atom);
+}
+
+void ofono_ussd_set_data(struct ofono_ussd *ussd, void *data)
+{
+	ussd->driver_data = data;
+}
+
+void *ofono_ussd_get_data(struct ofono_ussd *ussd)
+{
+	return ussd->driver_data;
+}
